//-----------------------------------------------------------------------------
// File: Meshes.cpp
//
// Desc: For advanced geometry, most apps will prefer to load pre-authored
//       Meshes from a file. Fortunately, when using Meshes, D3DX does most of
//       the work for this, parsing a geometry file and creating vertx buffers
//       (and index buffers) for us. This tutorial shows how to use a D3DXMESH
//       object, including loading it from a file and rendering it. One thing
//       D3DX does not handle for us is the materials and textures for a mesh,
//       so note that we have to handle those manually.
//
//       Note: one advanced (but nice) feature that we don't show here is that
//       when cloning a mesh we can specify the FVF. So, regardless of how the
//       mesh was authored, we can add/remove normals, add more texture
//       coordinate sets (for multi-texturing), etc.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
#include <Windows.h>
#include <mmsystem.h>
#include <d3dx9.h>
#pragma warning( disable : 4996 ) // disable deprecated warning 
#include <strsafe.h>
#pragma warning( default : 4996 ) 




//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
LPDIRECT3D9             g_pD3D           = NULL; // Used to create the D3DDevice
LPDIRECT3DDEVICE9       g_pd3dDevice     = NULL; // Our rendering device

LPD3DXMESH              g_pMesh          = NULL; // Our mesh object in sysmem
D3DMATERIAL9*           g_pMeshMaterials = NULL; // Materials for our mesh
LPDIRECT3DTEXTURE9*     g_pMeshTextures  = NULL; // Textures for our mesh
DWORD                   g_dwNumMaterials = 0L;   // Number of mesh materials


LPDIRECT3DSURFACE9 g_pDirect3DSurfaceBack = NULL;

LPDIRECT3DTEXTURE9 g_pDirect3DTextureFirst = NULL;
LPDIRECT3DSURFACE9 g_pDirect3DSurfaceFirst = NULL;
LPDIRECT3DTEXTURE9 g_pDirect3DTextureSecond = NULL;
LPDIRECT3DSURFACE9 g_pDirect3DSurfaceSecond = NULL;

LPDIRECT3DSURFACE9 g_pZStencilSurface = NULL;

struct VERTEX_XYZ_RHW_COLOR
{
	float x, y, z;
	float rhw;
	D3DCOLOR color;
};

#define	D3DFVF_XYZ_RHW_COLOR	( D3DFVF_XYZRHW | D3DFVF_DIFFUSE )

struct VERTEX_XYZ_RHW_COLOR_UV
{
	float x, y, z;
	float rhw;
	D3DCOLOR color;
	float u, v;
};

#define	D3DFVF_XYZ_RHW_COLOR_UV	( D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1 )

//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
HRESULT InitD3D( HWND hWnd )
{
    // Create the D3D object.
    if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
        return E_FAIL;

    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

	d3dpp.MultiSampleType = D3DMULTISAMPLE_NONE;
	d3dpp.BackBufferWidth  = 512;
	d3dpp.BackBufferHeight = 512;

    // Create the D3DDevice
    if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &d3dpp, &g_pd3dDevice ) ) )
    {
        return E_FAIL;
    }

    // Turn on the zbuffer
    g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );

    // Turn on ambient lighting 
    g_pd3dDevice->SetRenderState( D3DRS_AMBIENT, 0xffffffff );

	// AiOtp̃eNX`쐬
	g_pd3dDevice->GetRenderTarget( 0, &g_pDirect3DSurfaceBack );

	D3DSURFACE_DESC Desc;
	g_pDirect3DSurfaceBack->GetDesc( &Desc );

	g_pd3dDevice->CreateTexture( 512, 512, 1, D3DUSAGE_RENDERTARGET,
		Desc.Format, D3DPOOL_DEFAULT, &g_pDirect3DTextureFirst, NULL );

	g_pd3dDevice->CreateTexture( 512, 512, 1, D3DUSAGE_RENDERTARGET,
		Desc.Format, D3DPOOL_DEFAULT, &g_pDirect3DTextureSecond, NULL );

	g_pDirect3DTextureFirst->GetSurfaceLevel( 0, &g_pDirect3DSurfaceFirst );

	g_pDirect3DTextureSecond->GetSurfaceLevel( 0, &g_pDirect3DSurfaceSecond );

	g_pd3dDevice->CreateDepthStencilSurface( 512, 512,
		D3DFMT_D16, D3DMULTISAMPLE_NONE, 0, FALSE, &g_pZStencilSurface, NULL );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: InitGeometry()
// Desc: Load the mesh and build the material and texture arrays
//-----------------------------------------------------------------------------
HRESULT InitGeometry()
{
    LPD3DXBUFFER pD3DXMtrlBuffer;

    // Load the mesh from the specified file
    if( FAILED( D3DXLoadMeshFromX( "Tiger.x", D3DXMESH_SYSTEMMEM, 
                                   g_pd3dDevice, NULL, 
                                   &pD3DXMtrlBuffer, NULL, &g_dwNumMaterials, 
                                   &g_pMesh ) ) )
    {
        // If model is not in current folder, try parent folder
        if( FAILED( D3DXLoadMeshFromX( "..\\Tiger.x", D3DXMESH_SYSTEMMEM, 
                                    g_pd3dDevice, NULL, 
                                    &pD3DXMtrlBuffer, NULL, &g_dwNumMaterials, 
                                    &g_pMesh ) ) )
        {
            MessageBox(NULL, "Could not find tiger.x", "Meshes.exe", MB_OK);
            return E_FAIL;
        }
    }

    // We need to extract the material properties and texture names from the 
    // pD3DXMtrlBuffer
    D3DXMATERIAL* d3dxMaterials = (D3DXMATERIAL*)pD3DXMtrlBuffer->GetBufferPointer();
    g_pMeshMaterials = new D3DMATERIAL9[g_dwNumMaterials];
    if( g_pMeshMaterials == NULL )
        return E_OUTOFMEMORY;
    g_pMeshTextures  = new LPDIRECT3DTEXTURE9[g_dwNumMaterials];
    if( g_pMeshTextures == NULL )
        return E_OUTOFMEMORY;

    for( DWORD i=0; i<g_dwNumMaterials; i++ )
    {
        // Copy the material
        g_pMeshMaterials[i] = d3dxMaterials[i].MatD3D;

        // Set the ambient color for the material (D3DX does not do this)
        g_pMeshMaterials[i].Ambient = g_pMeshMaterials[i].Diffuse;

        g_pMeshTextures[i] = NULL;
        if( d3dxMaterials[i].pTextureFilename != NULL && 
            lstrlen(d3dxMaterials[i].pTextureFilename) > 0 )
        {
            // Create the texture
            if( FAILED( D3DXCreateTextureFromFile( g_pd3dDevice, 
                                                d3dxMaterials[i].pTextureFilename, 
                                                &g_pMeshTextures[i] ) ) )
            {
                // If texture is not in current folder, try parent folder
                const TCHAR* strPrefix = TEXT("..\\");
                TCHAR strTexture[MAX_PATH];
                StringCchCopy( strTexture, MAX_PATH, strPrefix );
                StringCchCat( strTexture, MAX_PATH, d3dxMaterials[i].pTextureFilename );
                // If texture is not in current folder, try parent folder
                if( FAILED( D3DXCreateTextureFromFile( g_pd3dDevice, 
                                                    strTexture, 
                                                    &g_pMeshTextures[i] ) ) )
                {
                    MessageBox(NULL, "Could not find texture map", "Meshes.exe", MB_OK);
                }
            }
        }
    }

    // Done with the material buffer
    pD3DXMtrlBuffer->Release();

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
VOID Cleanup()
{
	g_pZStencilSurface->Release();
	g_pDirect3DSurfaceSecond->Release();
	g_pDirect3DTextureSecond->Release();
	g_pDirect3DSurfaceFirst->Release();
	g_pDirect3DTextureFirst->Release();
	g_pDirect3DSurfaceBack->Release();

    if( g_pMeshMaterials != NULL ) 
        delete[] g_pMeshMaterials;

    if( g_pMeshTextures )
    {
        for( DWORD i = 0; i < g_dwNumMaterials; i++ )
        {
            if( g_pMeshTextures[i] )
                g_pMeshTextures[i]->Release();
        }
        delete[] g_pMeshTextures;
    }
    if( g_pMesh != NULL )
        g_pMesh->Release();
    
    if( g_pd3dDevice != NULL )
        g_pd3dDevice->Release();

    if( g_pD3D != NULL )
        g_pD3D->Release();
}



//-----------------------------------------------------------------------------
// Name: SetupMatrices()
// Desc: Sets up the world, view, and projection transform matrices.
//-----------------------------------------------------------------------------
VOID SetupMatrices()
{
    // Set up world matrix
    D3DXMATRIXA16 matWorld;
    D3DXMatrixRotationY( &matWorld, timeGetTime()/1000.0f );
    g_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

    // For the projection matrix, we set up a perspective transform (which
    // transforms geometry from 3D view space to 2D viewport space, with
    // a perspective divide making objects smaller in the distance). To build
    // a perpsective transform, we need the field of view (1/4 pi is common),
    // the aspect ratio, and the near and far clipping planes (which define at
    // what distances geometry should be no longer be rendered).
    D3DXMATRIXA16 matProj;
    D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, 1.0f, 1.0f, 100.0f );
    g_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );
}




//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
VOID Render()
{
    // Clear the backbuffer and the zbuffer
    g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 
                         D3DCOLOR_XRGB(0,0,0), 1.0f, 0 );
    
    // Begin the scene
    if( SUCCEEDED( g_pd3dDevice->BeginScene() ) )
    {
        // Setup the world, view, and projection matrices
        SetupMatrices();

		D3DXVECTOR3 vEyePt( -0.15f, 3.0f, -5.0f );
		D3DXVECTOR3 vLookatPt( 0.0f, 0.0f, 0.0f );
		D3DXVECTOR3 vUpVec( 0.0f, 1.0f, 0.0f );
		D3DXMATRIXA16 matView;
		D3DXMatrixLookAtLH( &matView, &vEyePt, &vLookatPt, &vUpVec );
		g_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );

		g_pd3dDevice->SetRenderTarget( 0, g_pDirect3DSurfaceFirst );

		g_pd3dDevice->SetDepthStencilSurface( g_pZStencilSurface );

		g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB( 0, 0, 0 ), 1, 0 );

		g_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
		g_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );

        // Meshes are divided into subsets, one for each material. Render them in
        // a loop
        for( DWORD i=0; i<g_dwNumMaterials; i++ )
        {
            // Set the material and texture for this subset
            g_pd3dDevice->SetMaterial( &g_pMeshMaterials[i] );
            g_pd3dDevice->SetTexture( 0, g_pMeshTextures[i] );
        
            // Draw the mesh subset
            g_pMesh->DrawSubset( i );
        }

		// VAɕϊ
		{
			VERTEX_XYZ_RHW_COLOR Vertex[4] = {
				{   0,   0, 0, 1.0f, D3DCOLOR_ARGB( 255, 0, 255, 255 ) },
				{ 512,   0, 0, 1.0f, D3DCOLOR_ARGB( 255, 0, 255, 255 ) },
				{ 512, 512, 0, 1.0f, D3DCOLOR_ARGB( 255, 0, 255, 255 ) },
				{   0, 512, 0, 1.0f, D3DCOLOR_ARGB( 255, 0, 255, 255 ) }
			};

			// eNX`ݒ
			g_pd3dDevice->SetTexture( 0, NULL );

			// _tH[}bg̐ݒ
			g_pd3dDevice->SetFVF( D3DFVF_XYZ_RHW_COLOR );

			g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
			g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
			g_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_ONE );
			g_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );

			g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLEFAN, 2, Vertex, sizeof( VERTEX_XYZ_RHW_COLOR ) );
		}

		vEyePt.x = 0.15f;
		D3DXMatrixLookAtLH( &matView, &vEyePt, &vLookatPt, &vUpVec );
		g_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );

		g_pd3dDevice->SetRenderTarget( 0, g_pDirect3DSurfaceSecond );
		g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB( 0, 0, 0 ), 1, 0 );

		g_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
		g_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );

        // Meshes are divided into subsets, one for each material. Render them in
        // a loop
        for( DWORD i=0; i<g_dwNumMaterials; i++ )
        {
            // Set the material and texture for this subset
            g_pd3dDevice->SetMaterial( &g_pMeshMaterials[i] );
            g_pd3dDevice->SetTexture( 0, g_pMeshTextures[i] );
        
            // Draw the mesh subset
            g_pMesh->DrawSubset( i );
        }

		// }[_+CG[ɕϊ
		{
			VERTEX_XYZ_RHW_COLOR Vertex[4] = {
				{   0,   0, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 0, 0 ) },
				{ 512,   0, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 0, 0 ) },
				{ 512, 512, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 0, 0 ) },
				{   0, 512, 0, 1.0f, D3DCOLOR_ARGB( 255, 255, 0, 0 ) }
			};

			// eNX`ݒ
			g_pd3dDevice->SetTexture( 0, NULL );

			// _tH[}bg̐ݒ
			g_pd3dDevice->SetFVF( D3DFVF_XYZ_RHW_COLOR );

			g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
			g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
			g_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_ONE );
			g_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );

			g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLEFAN, 2, Vertex, sizeof( VERTEX_XYZ_RHW_COLOR ) );
		}


		// obNobt@ɍځAEڕ`挋ʂ]
		g_pd3dDevice->SetRenderTarget( 0, g_pDirect3DSurfaceBack );

		VERTEX_XYZ_RHW_COLOR_UV Vertex[4] = {
			{   0,   0, 0, 1.0f, D3DCOLOR_ARGB( 128, 128, 128, 128 ), 0, 0 },
			{ 512,   0, 0, 1.0f, D3DCOLOR_ARGB( 128, 128, 128, 128 ), 1, 0 },
			{ 512, 512, 0, 1.0f, D3DCOLOR_ARGB( 128, 128, 128, 128 ), 1, 1 },
			{   0, 512, 0, 1.0f, D3DCOLOR_ARGB( 128, 128, 128, 128 ), 0, 1 }
		};

		// eNX`ݒ
		g_pd3dDevice->SetTexture( 0, g_pDirect3DTextureFirst );

		// _tH[}bg̐ݒ
		g_pd3dDevice->SetFVF( D3DFVF_XYZ_RHW_COLOR_UV );

		g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, FALSE );
		g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );

		g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLEFAN, 2, Vertex, sizeof( VERTEX_XYZ_RHW_COLOR_UV ) );

		VERTEX_XYZ_RHW_COLOR_UV Vertex2[4] = {
			{   0,   0, 0, 1.0f, D3DCOLOR_ARGB( 128, 128, 128, 128 ), 0, 0 },
			{ 512,   0, 0, 1.0f, D3DCOLOR_ARGB( 128, 128, 128, 128 ), 1, 0 },
			{ 512, 512, 0, 1.0f, D3DCOLOR_ARGB( 128, 128, 128, 128 ), 1, 1 },
			{   0, 512, 0, 1.0f, D3DCOLOR_ARGB( 128, 128, 128, 128 ), 0, 1 }
		};

		g_pd3dDevice->SetTexture( 0, g_pDirect3DTextureSecond );
		g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLEFAN, 2, Vertex2, sizeof( VERTEX_XYZ_RHW_COLOR_UV ) );

		g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );


        // End the scene
        g_pd3dDevice->EndScene();
    }

    // Present the backbuffer contents to the display
    g_pd3dDevice->Present( NULL, NULL, NULL, NULL );
}




//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
        case WM_DESTROY:
            Cleanup();
            PostQuitMessage( 0 );
            return 0;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}




//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    // Register the window class
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L, 
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      "D3D Tutorial", NULL };
    RegisterClassEx( &wc );

    // Create the application's window
    HWND hWnd = CreateWindow( "D3D Tutorial", "D3D Tutorial 06: Meshes", 
                              WS_OVERLAPPEDWINDOW, 100, 100, 512, 512,
                              NULL, NULL, wc.hInstance, NULL );

    // Initialize Direct3D
    if( SUCCEEDED( InitD3D( hWnd ) ) )
    { 
        // Create the scene geometry
        if( SUCCEEDED( InitGeometry() ) )
        {
            // Show the window
            ShowWindow( hWnd, SW_SHOWDEFAULT );
            UpdateWindow( hWnd );

            // Enter the message loop
            MSG msg; 
            ZeroMemory( &msg, sizeof(msg) );
            while( msg.message!=WM_QUIT )
            {
                if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
                {
                    TranslateMessage( &msg );
                    DispatchMessage( &msg );
                }
                else
                    Render();
            }
        }
    }

    UnregisterClass( "D3D Tutorial", wc.hInstance );
    return 0;
}



